name = "image_app"
version_info = (0, 0, 3, 20041619)
__version__ = ".".join([str(v) for v in version_info])
__description__ = '图片处理'

__all__ = ["load_smooth_area", 'cv2pil', 'pil2cv']

from PIL import Image
import cv2
import numpy as np


def load_smooth_area(arr_img):
    from .SmoothAreaProvider import SmoothAreaProvider
    return SmoothAreaProvider().get_image_rects(arr_img)


def cv2pil(arr_img):
    img = Image.fromarray(cv2.cvtColor(arr_img, cv2.COLOR_BGR2RGB))
    # img.show()
    return img


def pil2cv(img):
    assert isinstance(img, Image.Image)
    arr_img = cv2.cvtColor(np.asarray(img), cv2.COLOR_RGB2BGR)
    # cv2.imshow("OpenCV", arr_img)
    # cv2.waitKey()
    return arr_img



def select_by_kp(image_integral, anchors):
    pass


def nms(dets, thresh):
    pass


def box_connect(imageIntegral, anchors_in, h_gap, width, height):
    pass


from .select_by_kp import select_by_kp, nms, box_connect



import cv2
import math


def judge_similar_area(img, x1, y1, x2, y2):
    """
    判断区域(x1, y1, x2, y2)是否相似（是否轮廓突变，边界为轮廓突变）
    :param im: 背景图片
    :param x1:  区域起始列
    :param y1:  区域起始行
    :param x2:  区域结束
    :param y2:  区域结束行
    :return:    True:该区域为空白区域, False:该区域非空白区域
    """

    # img = cv2.cvtColor(np.asarray(im), cv2.COLOR_RGBA2BGRA)
    height, width = img.shape[:2]
    if x1 < 0 or y1 < 0 or x2 > width - 1 or y2 > height - 1 or x1 > x2 or y1 > y2:
        # 是否超出im范围
        return False

    crop = img[y1:y2, x1:x2]
    crop = cv2.Canny(crop, 64, 128)  # 检查阈值边缘差距
    contours, _ = cv2.findContours(crop, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    if len(contours) > 0:
        return False
    return True


def spreading_area(img, rect, align, offset, direct, limit=None):
    """
    对一个区域，向上下移动，找到安全的区域（区域不突变）
    :param rect:
    :param align:
    :param offset: 2
    :param direct: 0 y，1 x
    :param limit:
    :return:
    """
    if limit:
        limit = min([limit, img.shape[direct]])
    else:
        limit = img.shape[direct]
    index_1, index_2 = (0, 2) if direct else (1, 3)
    temp_rect = cur_rect = rect.copy()
    while True:
        is_similar_area = judge_similar_area(img, *temp_rect)
        if is_similar_area:
            cur_rect = temp_rect.copy()
            if align == 0:  # 做对齐
                ymax = temp_rect[index_2] + offset
                if ymax - temp_rect[index_1] + 1 > limit:
                    break
                temp_rect[index_2] = ymax
            elif align == 1:
                ymin, ymax = temp_rect[index_1] - offset, temp_rect[index_2] + offset
                if ymin < 0 or ymax - ymin + 1 > limit:
                    break
                temp_rect[index_1] = ymin
                temp_rect[index_2] = ymax
            else:
                ymin = temp_rect[index_1] - offset
                if ymin < 0 and temp_rect[index_2] - ymin + 1 > limit:
                    temp_rect[index_1] = ymin
                temp_rect[index_1] = ymin
        else:
            cur_rect[index_1] = max(cur_rect[index_1], 0)
            # cur_rect[index_2] = min(cur_rect[index_2], limit - 1)
            break
    return cur_rect


def multi_spreading_area(img, rect, factor_info):
    temp_rect = rect.copy()
    while True:
        cur_rect = temp_rect.copy()
        for index, factor in factor_info.items():
            temp_rect[index] += factor * (1 if index >= 2 else -1)
        is_similar_area = judge_similar_area(img, *rect)
        if not is_similar_area:
            break
    return cur_rect


def par_relarge_area(img, rect, xy_align, par):
    x_align, y_align = xy_align
    x_center, y_center = rect[0] + rect[2], rect[1] + rect[3]
    temp_rect = cur_rect = rect.copy()
    while True:
        is_similar_area = judge_similar_area(img, *temp_rect)
        if not is_similar_area:
            break

        cur_rect = temp_rect.copy()
        if par > 1:
            if y_align == 0:
                temp_rect[3] += 2
            elif y_align == 1:
                temp_rect[1] -= 1
                temp_rect[3] += 1
            else:
                temp_rect[1] -= 2

            req = math.ceil((temp_rect[3] - temp_rect[1] + 1) * par)
            if x_align == 0:
                temp_rect[2] = temp_rect[0] + req - 1
            elif x_align == 1:
                temp_rect[0] = int((x_center - req) / 2) + 1
                temp_rect[2] = req + temp_rect[0] - 1
            else:
                temp_rect[0] = temp_rect[2] - req + 1
        else:
            if x_align == 0:
                temp_rect[2] += 2
            elif x_align == 1:
                temp_rect[0] -= 1
                temp_rect[2] += 1
            else:
                temp_rect[0] -= 2

            req = math.ceil((temp_rect[2] - temp_rect[0] + 1) * par)
            if y_align == 0:
                temp_rect[3] = temp_rect[1] + req - 1
            elif y_align == 1:
                temp_rect[1] = int((y_center - req) / 2)
                temp_rect[3] = req + temp_rect[1] - 1
            else:
                temp_rect[1] = temp_rect[3] - req + 1

    return cur_rect




# cmake -D CMAKE_BUILD_TYPE=RELEASE -D CMAKE_INSTALL_PREFIX=/data/User/hanat/opencv_install -D WITH_CUDA=ON -D ENABLE_FAST_MATH=1 -D CUDA_FAST_MATH=1 -D WITH_CUBLAS=1 -D WITH_OPENMP=1 ..

import cv2
import numpy as np
from .libs import nms
from .libs import select_by_kp
from .libs import box_connect


def _whctrs(anchor):
    """
    Return width, height, x center, and y center for an anchor (window).
    """
    w = anchor[2] - anchor[0] + 1
    h = anchor[3] - anchor[1] + 1
    x_ctr = anchor[0] + 0.5 * (w - 1)
    y_ctr = anchor[1] + 0.5 * (h - 1)
    return w, h, x_ctr, y_ctr


def _ratio_enum(anchor, ratios):
    """
    Enumerate a set of anchors for each aspect ratio wrt an anchor.
    """

    w, h, x_ctr, y_ctr = _whctrs(anchor)
    size = w * h
    size_ratios = size / ratios
    ws = np.round(np.sqrt(size_ratios))
    hs = np.round(ws * ratios)
    anchors = _mkanchors(ws, hs, x_ctr, y_ctr)
    return anchors


def _scale_enum(anchor, scales):
    """
    Enumerate a set of anchors for each scale wrt an anchor.
    """
    w, h, x_ctr, y_ctr = _whctrs(anchor)
    ws = w * scales
    hs = h * scales
    anchors = _mkanchors(ws, hs, x_ctr, y_ctr)
    return anchors


def _mkanchors(ws, hs, x_ctr, y_ctr):
    """
    Given a vector of widths (ws) and heights (hs) around a center
    (x_ctr, y_ctr), output a set of anchors (windows).
    """
    ws = ws[:, np.newaxis]
    hs = hs[:, np.newaxis]
    anchors = np.hstack((x_ctr - 0.5 * (ws - 1),
                         y_ctr - 0.5 * (hs - 1),
                         x_ctr + 0.5 * (ws - 1),
                         y_ctr + 0.5 * (hs - 1),))
    return anchors


class SmoothAreaProvider(object):
    """

    """

    def __init__(self,
                 down_scale=32,
                 anchor_ratio=(0.17, 0.25, 0.5, 1.0, 2.0, 4.0, 6),
                 anchor_scale=(8, 16, 24, 32, 48, 64, 72, 90)
                 ):
        self._down_scale = down_scale
        self._anchor_ratio = eval(anchor_ratio) if type(anchor_ratio) == str else anchor_ratio
        self._anchor_scale = eval(anchor_scale) if type(anchor_scale) == str else anchor_scale
        # self._fast = cv2.FastFeatureDetector_create()
        self._fast = cv2.ORB_create(nfeatures=3000)

    def get_image_rects(self, image_in, long_side=320):
        """
        get region rects from image smoothness area
        :param image_in:
        :param long_side:
        :return:
        """
        assert type(image_in) == np.ndarray
        if len(image_in.shape) == 2:
            image_process = np.asarray(image_in)
        elif image_in.shape[2] == 3:
            image_process = cv2.cvtColor(image_in, cv2.COLOR_RGB2GRAY)
        elif image_in.shape[2] == 4:
            image_process = cv2.cvtColor(image_in, cv2.COLOR_RGBA2RGB)
        else:
            raise ValueError("wrong image format")

        src_h, src_w = image_process.shape[:2]
        if src_h > src_w:
            new_h = int(long_side)
            scale = src_h / long_side
            new_w = int(src_w / scale)
        else:
            new_w = int(long_side)
            scale = src_w / long_side
            new_h = int(src_h / scale)

        image_process = cv2.resize(image_process, (new_w, new_h))
        anchors, length = self.generate_anchors_pre(new_h, new_w)

        # key_points = self._fast.detect(image_process, None)
        kp_image = cv2.Canny(image_process, 50, 150) // 255
        sum_arr = np.zeros((new_h, new_w), np.float32)
        image_integral = cv2.integral(kp_image, sum_arr, cv2.CV_32FC1)
        anchor_preserved = select_by_kp(image_integral, anchors)
        length = len(anchor_preserved)
        anchor_preserved = np.asarray(anchor_preserved, np.float32)
        if length > 0:
            keep = nms(anchor_preserved, 0.00001)
            anchor_preserved = anchor_preserved[keep, :]
            anchor_preserved = box_connect(image_integral, anchor_preserved, 30, new_w, new_h)
            keep = nms(anchor_preserved, 0.00001)
            anchor_preserved = anchor_preserved[keep, :]

        for i in range(anchor_preserved.shape[0]):
            anchor_preserved[i, :] = anchor_preserved[i, :] * scale
        if anchor_preserved.shape[0] > 0:
            anchor_out = anchor_preserved[:, 0:4]
            anchor_out = np.asarray(anchor_out, np.int32)
            anchor_out = list(anchor_out)
            return anchor_out
        else:
            return list()

    def generate_anchors_pre(self, height, width):
        """ A wrapper function to generate anchors given different scales
          Also return the number of anchors in variable 'length'
        """
        anchors = self.generate_anchors(ratios=np.asarray(self._anchor_ratio), scales=np.asarray(self._anchor_scale))
        A = anchors.shape[0]
        shift_x = np.arange(0, width)  # * self._down_scale
        shift_y = np.arange(0, height)  # * self._down_scale
        shift_x, shift_y = np.meshgrid(shift_x, shift_y)
        shifts = np.vstack((shift_x.ravel(), shift_y.ravel(), shift_x.ravel(), shift_y.ravel())).transpose()
        K = shifts.shape[0]
        # width changes faster, so here it is H, W, C
        anchors = anchors.reshape((1, A, 4)) + shifts.reshape((1, K, 4)).transpose((1, 0, 2))
        anchors = anchors.reshape((K * A, 4)).astype(np.float32, copy=False)
        length = np.int32(anchors.shape[0])

        return anchors, length

    @staticmethod
    def generate_anchors(base_size=3, ratios=np.asarray([0.5, 1, 2]),
                         scales=2 ** np.arange(3, 6)):
        """
        Generate anchor (reference) windows by enumerating aspect ratios X
        scales wrt a reference (0, 0, 15, 15) window.
        """

        base_anchor = np.array([1, 1, base_size, base_size], np.float32) - 1
        ratio_anchors = _ratio_enum(base_anchor, ratios)
        anchors = np.vstack([_scale_enum(ratio_anchors[i, :], scales)
                             for i in range(ratio_anchors.shape[0])])
        return anchors


if __name__ == '__main__':
    import os
    path = os.path.abspath("1.jpg")
    smooth = SmoothAreaProvider()
    image = cv2.imread(path)
    rects = smooth.get_image_rects(image)
    for rect in rects:
        cv2.rectangle(image, (rect[0], rect[1]), (rect[2], rect[3]), (120, 78, 255), 2)
    cv2.imwrite('test.jpg', image)
    cv2.imshow("test", image)
    cv2.waitKey(3000)



# from distutils.core import setup, Extension
import numpy as np
from Cython.Build import cythonize
from Cython.Distutils import build_ext

numpy_include = np.get_include()

import os
from setuptools import setup, find_packages, Extension
from image_app import name, __description__, __version__

# read dev requirements
fname = os.path.join(os.path.dirname(__file__), 'requirements.txt')
with open(fname) as f:
    install_requires = [l.strip() for l in f.readlines()]
root = os.path.basename(os.path.dirname(os.path.abspath(__file__)))

with open("MAINTAINERS") as f:
    lines = [l.strip() for l in f.readlines()]
    author, author_email = lines[0].split(sep=" ")
    if len(lines) > 1:
        maintainer, maintainer_email = lines[1].split(sep=" ")
    else:
        maintainer, maintainer_email = author, author_email

try:
    import pypandoc

    long_description = pypandoc.convert('README.md', 'rst')
except:
    long_description = open('README.rst', encoding='utf-8').read()

setup(
    name=name,
    # cmdclass={'build_ext': build_ext},
    # ext_modules=cythonize(
    #     [Extension('image_app/libs/select_by_kp', ["src/select_by_kp.pyx"], include_dirs=[numpy_include, ])]
    # ),
    version=__version__,
    packages=find_packages(exclude=['examples', 'tests']),
    url='https://github.com/BingerYang/{}'.format(root),
    license='MITT',
    author=author,
    author_email=author_email,
    maintainer=maintainer,
    maintainer_email=maintainer_email,
    description=__description__,
    long_description=long_description,
    long_description_content_type="text/x-rst",

    python_requires='>=3.4',
    platforms=['all'],
    include_package_data=True,
    package_data={'src': ["*.c"]},
)



import numpy as np
cimport numpy as np
from libc.stdio cimport printf
# from cython.parallel import prange, parallel, threadid

cdef np.float32_t sigmoid(np.float32_t x):
    return 1.0 / (1.0 + np.exp(-x))

cdef inline np.float32_t max(np.float32_t a, np.float32_t b):
    return a if a >= b else b

cdef inline np.float32_t min(np.float32_t a, np.float32_t b):
    return a if a <= b else b

def select_by_kp(np.ndarray[np.float32_t, ndim=2] imageIntegral, np.ndarray[np.float32_t, ndim=2] anchors):
    """

    :param imageIntegral:
    :param anchors:
    :return:
    """

    anchor_preserved = []
    # score_preserved = []
    cdef int length = anchors.shape[0]
    cdef int count = 0
    cdef int x1, y1, x2, y2,
    cdef int w = imageIntegral.shape[1] - 1
    cdef int h = imageIntegral.shape[0] - 1
    cdef float image_area = float(w * h)
    cdef float area, integral = 0.0
    cdef int add = 0;
    while count < length:
        x1 = int(min(max(0, anchors[count, 0]), w))
        y1 = int(min(max(0, anchors[count, 1]), h))
        x2 = int(min(max(0, anchors[count, 2]), w))
        y2 = int(min(max(0, anchors[count, 3]), h))
        integral = imageIntegral[y2, x2] - imageIntegral[y2, x1] - imageIntegral[y1, x2] + imageIntegral[y1, x1]
        if integral < 1.0:
            area = float((y2 - y1) * (x2 - x1)) / float(image_area)
            # select_out[add, :] = [x1, y1, x2, y2, area]
            anchor_preserved.append([x1, y1, x2, y2, area])
        count += 1
        integral = 0.0
    return anchor_preserved


def nms(np.ndarray[np.float32_t, ndim=2] dets, np.float thresh):
    """

    :param dets:
    :param thresh:
    :return:
    """
    cdef np.ndarray[np.float32_t, ndim=1] x1 = dets[:, 0]
    cdef np.ndarray[np.float32_t, ndim=1] y1 = dets[:, 1]
    cdef np.ndarray[np.float32_t, ndim=1] x2 = dets[:, 2]
    cdef np.ndarray[np.float32_t, ndim=1] y2 = dets[:, 3]
    cdef np.ndarray[np.float32_t, ndim=1] scores = dets[:, 4]

    cdef np.ndarray[np.float32_t, ndim=1] areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    cdef np.ndarray[np.int_t, ndim=1] order = scores.argsort()[::-1]

    cdef int ndets = dets.shape[0]
    cdef np.ndarray[np.int_t, ndim=1] suppressed = \
        np.zeros((ndets), dtype=np.int)

    # nominal indices
    cdef int _i, _j
    # sorted indices
    cdef int i, j
    # temp variables for box i's (the box currently under consideration)
    cdef np.float32_t ix1, iy1, ix2, iy2, iarea
    # variables for computing overlap with box j (lower scoring box)
    cdef np.float32_t xx1, yy1, xx2, yy2
    cdef np.float32_t w, h
    cdef np.float32_t inter, ovr

    keep = []
    for _i in range(ndets):
        i = order[_i]
        if suppressed[i] == 1:
            continue
        keep.append(i)
        ix1 = x1[i]
        iy1 = y1[i]
        ix2 = x2[i]
        iy2 = y2[i]
        iarea = areas[i]
        for _j in range(_i + 1, ndets):
            j = order[_j]
            if suppressed[j] == 1:
                continue
            xx1 = max(ix1, x1[j])
            yy1 = max(iy1, y1[j])
            xx2 = min(ix2, x2[j])
            yy2 = min(iy2, y2[j])
            w = max(0.0, xx2 - xx1 + 1)
            h = max(0.0, yy2 - yy1 + 1)
            inter = w * h
            ovr = inter / (iarea + areas[j] - inter)
            if ovr >= thresh:
                suppressed[j] = 1

    return keep


cdef inline np.int vertical_similarity(np.ndarray[np.float32_t, ndim=1] box1, np.ndarray[np.float32_t, ndim=1] box2, np.float iou_thresh):

    cdef float intersec_height = min(box1[3], box2[3]) - max(box1[1], box2[1])
    cdef float union_height = max(box1[3], box2[3]) - min(box1[1], box2[1])
    cdef float iou = intersec_height / union_height
    if iou > iou_thresh:
        return 1
    else:
        return 0####


def box_connect(np.ndarray[np.float32_t, ndim=2] imageIntegral, np.ndarray[np.float32_t, ndim=2] anchors_in, np.int h_gap, np.int width, np.int height):
    """

    :param imageIntegral:
    :param anchor:
    :return:
    """

    cdef int length = anchors_in.shape[0]
    cdef np.ndarray[np.float32_t, ndim=2] anchors = anchors_in[:, 0:4]
    cdef int x1, y1, x2, y2
    cdef float area, original_area, integral=0.0
    cdef int iou
    cdef float image_area = float(width * height)
    # left to right search
    for box_index1 in range(length):
        box1 = anchors[box_index1, :]
        for box_index2 in range(length):
            box2 = anchors[box_index2, :]
            iou = vertical_similarity(box1, box2, 0.6)
            if box1[2] <= box2[0] <= (box1[2] + h_gap) and iou==1:
                x1 = box1[0]
                y1 = int(min(box1[1], box2[1]))
                x2 = box2[2]
                y2 = int(max(box1[3], box2[3]))
                integral = imageIntegral[y2, x2] - imageIntegral[y2, x1] - imageIntegral[y1, x2] + imageIntegral[y1, x1]
                area = float((y2 - y1) * (x2 - x1))
                if integral<1:
                    area = area / image_area
                    anchors_in[box_index1, 0] = float(x1)
                    anchors_in[box_index1, 1] = float(y1)
                    anchors_in[box_index1, 2] = float(x2)
                    anchors_in[box_index1, 3] = float(y2)
                    anchors_in[box_index1, 4] = area
                    continue

                x1 = box1[0]
                y1 = int(min(box1[1], box2[1]))
                x2 = box2[2]
                y2 = int(min(box1[3], box2[3]))
                integral = imageIntegral[y2, x2] - imageIntegral[y2, x1] - imageIntegral[y1, x2] + imageIntegral[y1, x1]
                area = float((y2 - y1) * (x2 - x1))
                original_area = float((box1[2]-box1[0])*(box1[3]-box1[1])+ (box2[2]-box2[0])*(box2[3]-box2[1]))
                if integral<1 and area > original_area:
                    area = area / image_area
                    anchors_in[box_index1, 0] = float(x1)
                    anchors_in[box_index1, 1] = float(y1)
                    anchors_in[box_index1, 2] = float(x2)
                    anchors_in[box_index1, 3] = float(y2)
                    anchors_in[box_index1, 4] = area
                    continue

                x1 = box1[0]
                y1 = int(max(box1[1], box2[1]))
                x2 = box2[2]
                y2 = int(max(box1[3], box2[3]))
                integral = imageIntegral[y2, x2] - imageIntegral[y2, x1] - imageIntegral[y1, x2] + imageIntegral[y1, x1]
                area = float((y2 - y1) * (x2 - x1))
                original_area = float((box1[2]-box1[0])*(box1[3]-box1[1])+ (box2[2]-box2[0])*(box2[3]-box2[1]))
                if integral<1 and area > original_area:
                    area = area / image_area
                    anchors_in[box_index1, 0] = float(x1)
                    anchors_in[box_index1, 1] = float(y1)
                    anchors_in[box_index1, 2] = float(x2)
                    anchors_in[box_index1, 3] = float(y2)
                    anchors_in[box_index1, 4] = area
                    continue

    return anchors_in